@@ -20,7 +20,7 @@ from functools import partial
 from threading import Lock
 from subprocess import Popen, PIPE
 import cStringIO
-#from traceback import print_stack
+import string
 
 __all__ = ["git_version", "GitError", "GitErrorSha", "Storage", "StorageFactory"]
 
@@ -51,25 +51,31 @@ class GitCore:
 
         #print >>sys.stderr, "DEBUG:", git_cmd, cmd_args
 
-        p = Popen(self.__build_git_cmd(git_cmd, *cmd_args),
-                  stdin=None, stdout=PIPE, stderr=PIPE, close_fds=True)
+        if sys.platform == "win32":
+            p = Popen(self.__build_git_cmd(git_cmd, *cmd_args),
+                      stdin=None, stdout=PIPE, stderr=PIPE)
+        else:
+            p = Popen(self.__build_git_cmd(git_cmd, *cmd_args),
+                      stdin=None, stdout=PIPE, stderr=PIPE, close_fds=True)
 
         stdout_data, stderr_data = p.communicate()
         #TODO, do something with p.returncode, e.g. raise exception
 
-        return cStringIO.StringIO(stdout_data)
+        return stdout_data
 
     def __getattr__(self, name):
         return partial(self.__execute, name.replace('_','-'))
 
-    @staticmethod
-    def is_sha(sha):
+    __is_sha_pat = re.compile(r'[0-9A-Fa-f]*$')
+
+    @classmethod
+    def is_sha(cls,sha):
         """returns whether sha is a potential sha id
         (i.e. proper hexstring between 4 and 40 characters"""
-        if len(sha) < 4 or len(sha) > 40:
+        if not (4 <= len(sha) <= 40):
             return False
-        HEXCHARS = "0123456789abcdefABCDEF"
-        return all(s in HEXCHARS for s in sha)
+
+        return bool(cls.__is_sha_pat.match(sha))
 
 # helper class for caching...
 class SizedDict(dict):
@@ -145,11 +151,10 @@ class Storage:
 
     @staticmethod
     def git_version(git_bin="git"):
-        GIT_VERSION_MIN_REQUIRED = (1,5,2)
+        GIT_VERSION_MIN_REQUIRED = (1,5,6)
         try:
             g = GitCore(git_bin=git_bin)
-            output = g.version()
-            [v] = output.readlines()
+            [v] = g.version().splitlines()
             _,_,version = v.strip().split()
             # 'version' has usually at least 3 numeric version components, e.g.::
             #  1.5.4.2
@@ -202,9 +207,7 @@ class Storage:
         self.__commit_msg_cache = SizedDict(200)
         self.__commit_msg_lock = Lock()
 
-        # cache the last 2000 file sizes
-        self.__fs_obj_size_cache = SizedDict(2000)
-        self.__fs_obj_size_lock = Lock()
+
 
     def __del__(self):
         self.logger.debug("PyGIT.Storage instance %d destructed" % id(self))
@@ -240,7 +243,7 @@ class Storage:
                 new_tags = set([])
                 youngest = None
                 oldest = None
-                for revs in self.repo.rev_parse("--tags"):
+                for revs in self.repo.rev_parse("--tags").splitlines():
                     new_tags.add(revs.strip())
 
                 # helper for reusing strings
@@ -250,7 +253,7 @@ class Storage:
                     return __rev_seen.setdefault(rev, rev)
 
                 rev = ord_rev = 0
-                for revs in self.repo.rev_list("--parents", "--all"):
+                for revs in self.repo.rev_list("--parents", "--all").splitlines():
                     revs = revs.strip().split()
 
                     revs = map(__rev_reuse, revs)
@@ -372,7 +375,7 @@ class Storage:
     def get_commit_encoding(self):
         if self.commit_encoding is None:
             self.commit_encoding = \
-                self.repo.repo_config("--get", "i18n.commitEncoding").read().strip() or 'utf-8'
+                self.repo.repo_config("--get", "i18n.commitEncoding").strip() or 'utf-8'
 
         return self.commit_encoding
 
@@ -393,7 +396,7 @@ class Storage:
                 return fullrev
 
         # fall back to external git calls
-        rc = self.repo.rev_parse("--verify", rev).read().strip()
+        rc = self.repo.rev_parse("--verify", rev).strip()
         if not rc:
             return None
 
@@ -401,7 +404,7 @@ class Storage:
             return rc
 
         if rc in tag_db:
-            sha=self.repo.cat_file("tag", rc).read().split(None, 2)[:2]
+            sha=self.repo.cat_file("tag", rc).split(None, 2)[:2]
             if sha[0] != 'object':
                 self.logger.debug("unexpected result from 'git-cat-file tag %s'" % rc)
                 return None
@@ -412,7 +415,7 @@ class Storage:
     def shortrev(self, rev, min_len=7):
         "try to shorten sha id"
         #try to emulate the following:
-        #return self.repo.rev_parse("--short", str(rev)).read().strip()
+        #return self.repo.rev_parse("--short", str(rev)).strip()
         rev = str(rev)
 
         if min_len < self.__SREV_MIN:
@@ -466,7 +469,7 @@ class Storage:
     def get_branches(self):
         "returns list of (local) branches, with active (= HEAD) one being the first item"
         result=[]
-        for e in self.repo.branch("-v", "--no-abbrev"):
+        for e in self.repo.branch("-v", "--no-abbrev").splitlines():
             (bname,bsha)=e[1:].strip().split()[:2]
             if e.startswith('*'):
                 result.insert(0,(bname,bsha))
@@ -475,35 +478,39 @@ class Storage:
         return result
 
     def get_tags(self):
-        return [e.strip() for e in self.repo.tag("-l")]
+        return [e.strip() for e in self.repo.tag("-l").splitlines()]
 
     def ls_tree(self, rev, path=""):
-        rev = str(rev) # paranoia
+        rev = rev and str(rev) or 'HEAD' # paranoia
         if path.startswith('/'):
             path = path[1:]
 
-        if path:
-            tree = self.repo.ls_tree("-z", rev, "--", path)
-        else:
-            tree = self.repo.ls_tree("-z", rev)
+        tree = self.repo.ls_tree("-z", "-l", rev, "--", path).split('\0')
 
         def split_ls_tree_line(l):
-            "split according to '<mode> <type> <sha>\t<fname>'"
+            "split according to '<mode> <type> <sha> <size>\t<fname>'"
             meta,fname = l.split('\t')
-            _mode,_type,_sha = meta.split(' ')
-            return _mode,_type,_sha,fname
+            _mode,_type,_sha,_size = meta.split()
 
-        return [split_ls_tree_line(e) for e in tree.read().split('\0') if e]
+            if _size == '-':
+                _size = None
+            else:
+                _size = int(_size)
+
+            return _mode,_type,_sha,_size,fname
+
+        return [ split_ls_tree_line(e) for e in tree if e ]
 
     def read_commit(self, commit_id):
         if not commit_id:
             raise GitError("read_commit called with empty commit_id")
 
-        commit_id = str(commit_id)
+        commit_id, commit_id_orig = self.fullrev(commit_id), commit_id
 
         db = self.get_commits()
         if commit_id not in db:
-            self.logger.info("read_commit failed for '%s'" % commit_id)
+            self.logger.info("read_commit failed for '%s' ('%s')" %
+                             (commit_id, commit_id_orig))
             raise GitErrorSha
 
         with self.__commit_msg_lock:
@@ -513,7 +520,7 @@ class Storage:
                 return result[0], dict(result[1])
 
             # cache miss
-            raw = self.repo.cat_file("commit", commit_id).read()
+            raw = self.repo.cat_file("commit", commit_id)
             raw = unicode(raw, self.get_commit_encoding(), 'replace')
             lines = raw.splitlines()
 
@@ -534,17 +541,13 @@ class Storage:
             return result[0], dict(result[1])
 
     def get_file(self, sha):
-        return self.repo.cat_file("blob", str(sha))
+        return cStringIO.StringIO(self.repo.cat_file("blob", str(sha)))
 
     def get_obj_size(self, sha):
         sha = str(sha)
+
         try:
-            with self.__fs_obj_size_lock:
-                if self.__fs_obj_size_cache.has_key(sha):
-                    obj_size = self.__fs_obj_size_cache[sha]
-                else:
-                    obj_size = int(self.repo.cat_file("-s", sha).read().strip())
-                    self.__fs_obj_size_cache[sha] = obj_size
+            obj_size = int(self.repo.cat_file("-s", sha).strip())
         except ValueError:
             raise GitErrorSha("object '%s' not found" % sha)
 
@@ -590,26 +593,26 @@ class Storage:
         return self.get_commits().iterkeys()
 
     def sync(self):
-        rev = self.repo.rev_list("--max-count=1", "--all").read().strip()
+        rev = self.repo.rev_list("--max-count=1", "--all").strip()
         return self.__rev_cache_sync(rev)
 
     def last_change(self, sha, path):
         return self.repo.rev_list("--max-count=1",
-                                  sha, "--", path).read().strip() or None
+                                  sha, "--", path).strip() or None
 
     def history(self, sha, path, limit=None):
         if limit is None:
             limit = -1
-        for rev in self.repo.rev_list("--max-count=%d" % limit,
-                                      str(sha), "--", path):
-            yield rev.strip()
+
+        tmp = self.repo.rev_list("--max-count=%d" % limit, str(sha), "--", path)
+        return [ rev.strip() for rev in tmp.splitlines() ]
 
     def history_timerange(self, start, stop):
-        for rev in self.repo.rev_list("--reverse",
-                                      "--max-age=%d" % start,
-                                      "--min-age=%d" % stop,
-                                      "--all"):
-            yield rev.strip()
+        return [ rev.strip() for rev in \
+                     self.repo.rev_list("--reverse",
+                                        "--max-age=%d" % start,
+                                        "--min-age=%d" % stop,
+                                        "--all").splitlines() ]
 
     def rev_is_anchestor_of(self, rev1, rev2):
         """return True if rev2 is successor of rev1"""
@@ -620,7 +623,7 @@ class Storage:
     def blame(self, commit_sha, path):
         in_metadata = False
 
-        for line in self.repo.blame("-p", "--", path, str(commit_sha)):
+        for line in self.repo.blame("-p", "--", path, str(commit_sha)).splitlines():
             assert line
             if in_metadata:
                 in_metadata = not line.startswith('\t')
@@ -644,6 +647,7 @@ class Storage:
         # diff-tree returns records with the following structure:
         # :<old-mode> <new-mode> <old-sha> <new-sha> <change> NUL <old-path> NUL [ <new-path> NUL ]
 
+        path = path.strip("/")
         diff_tree_args = ["-z", "-r"]
         if find_renames:
             diff_tree_args.append("-M")
@@ -651,7 +655,7 @@ class Storage:
                                str(tree2),
                                "--", path])
 
-        lines = self.repo.diff_tree(*diff_tree_args).read().split('\0')
+        lines = self.repo.diff_tree(*diff_tree_args).split('\0')
 
         assert lines[-1] == ""
         del lines[-1]
@@ -690,6 +694,12 @@ class Storage:
 if __name__ == '__main__':
     import sys, logging, timeit
 
+    assert not GitCore.is_sha("123")
+    assert GitCore.is_sha("1a3f")
+    assert GitCore.is_sha("f"*40)
+    assert not GitCore.is_sha("x"+"f"*39)
+    assert not GitCore.is_sha("f"*41)
+
     print "git version [%s]" % str(Storage.git_version())
 
     # custom linux hack reading `/proc/<PID>/statm`
@@ -717,10 +727,10 @@ if __name__ == '__main__':
     __data_size_last = __data_size
 
     def print_data_usage():
-	global __data_size_last
-	__tmp = proc_statm()[5]
+        global __data_size_last
+        __tmp = proc_statm()[5]
         print "DATA: %6d %+6d" % (__tmp - __data_size, __tmp - __data_size_last)
-	__data_size_last = __tmp
+        __data_size_last = __tmp
 
     print_data_usage()
 
@@ -763,8 +773,6 @@ if __name__ == '__main__':
         for sha in g.children_recursive(head):
             if sha in seen:
                 print "dupe detected :-/", sha, len(seen)
-                #print seen
-                #break
             seen.add(sha)
         return seen
 
@@ -795,10 +803,10 @@ if __name__ == '__main__':
 
     # perform typical trac operations:
 
-    if 0:
+    if 1:
         print "--------------"
         rev = g.head()
-        for mode,type,sha,name in g.ls_tree(rev):
+        for mode,type,sha,_size,name in g.ls_tree(rev):
             [last_rev] = g.history(rev, name, limit=1)
             s = g.get_obj_size(sha) if type == "blob" else 0
             msg = g.read_commit(last_rev)
